/*
   use the program as you want but on your own risc and only if you 
   understand its purpose. 
   this may work only under plain DOS compiled with Turbo C 3.0
*/
#include <stdio.h>   /* printf()     */
#include <io.h>      /* filelength() */
#include <fcntl.h>   /* open()       */
#include <stdlib.h>  /* exit()       */
#include <ctype.h>   /* strtol()     */
#include <string.h>  /* strcpy()     */
#include <conio.h>   /* getch()      */
#include <bios.h>
/*****************************************************************************
 the following defines the default params they are choosen based on the
 original hdd. Exept for the SECTORSIZE all of them can be overridden by the
 comandline
*****************************************************************************/
#define SECTORSIZE      512L   /* sector containes 512 bytes */
#define MAXCYLINDER     980L   /* CHS parameter from the original hdd */
#define MAXHEAD          15L
#define MAXSECTOR        17L
#define TARGETDISK      0x81   /* 0x80 first hdd 0x81 second hdd ... */

#define _DISK_READ   2      /* AH function codes */
#define _DISK_WRITE  3
#define _DISK_INFO   8
typedef struct _tCHS
{
   struct
   {
      long  Cylinder;  /* current chs values */
      long  Head;
      long  Sector;
   } cur;
   struct 
   {
      long  Cylinder;    /* maximun chs values */
      long  Head;
      long  Sector;
   } max;   
}tCHS;

char  WrBuffer[SECTORSIZE];  /* just big enough for one sector */
char  RdBuffer[SECTORSIZE]; 
int   Test = 1;

/** calculate the CHS parameter based on a given sector number */
int MakeChs(tCHS *chs,long number)
{
   long remain,sector,head,cylinder;

   remain   = number;

   cylinder = remain   / (chs->max.Head * chs->max.Sector);
   remain  -= cylinder * (chs->max.Head * chs->max.Sector);

   head     = remain / chs->max.Sector;
   remain  -= head   * chs->max.Sector;

   sector   = remain;

   if ( (head     >= chs->max.Head)   || 
        (sector   >= chs->max.Sector) || 
        (cylinder >= chs->max.Cylinder) )
   {  /* overflow occured */
      return 0;
   }

   /* report the parameter back */
   chs->cur.Cylinder = cylinder;
   chs->cur.Head     = head;
   chs->cur.Sector   = sector+1;
   return 1;
}

/** report parts of a buffer */ 
void PrintSector(char *buffer,int lines)
{
   int i,j;
   for (i=0;i<lines;i++)
   {
      printf("%4.4x : ",i*0x10);
      for (j=0;j<16;j++) printf("%2.2x ",buffer[j+i*0x10]);
      for (j=0;j<16;j++) printf("%c",    buffer[j+i*0x10]);
      printf ("\n");
   }
}

/** compare two buffer to verify the contents */
int CompareBuffer (char *buffer1, char *buffer2, int size)
{
   int i;
   for (i=0;i<size;i++)
   {
      if (*buffer1++ != *buffer2++) return i; /* error pos*/
   }
   return -1; /* meaning contents are the same*/
}


/** write one sector using INT13 subfunction 3 */
int WriteSector(tCHS *chs, char *buffer,int diskno)
{
   int result;

   int cyl  = chs->cur.Cylinder;
   int head = chs->cur.Head;
   int sect = chs->cur.Sector;

   result = biosdisk ( _DISK_WRITE,     /* Int13 AH== 0x3 write */
                       diskno,          /* drive 1 == 0x81 */
                       head,cyl,sect,   /* chs params */
                       1, buffer);      /* 1 sector */

   if ((result & 0xFF) ==0x11)  /* set <CRC corrected> to no error */
      result = 0;
   return result;
}

/** read one sector using INT13 subfunction 2 */
int ReadSector(tCHS *chs,char *buffer,int diskno)
{
   int result;

   int cyl  = chs->cur.Cylinder;
   int head = chs->cur.Head;
   int sect = chs->cur.Sector;

   result = biosdisk ( _DISK_READ,      /* Int13 AH== 0x2 read */
                       diskno,          /* drive 1 == 0x81 */
                       head,cyl,sect,   /* chs params */
                       1, buffer);      /* 1 sector */

   if ((result & 0xFF) ==0x11)  /* set <CRC corrected> to no error */
      result = 0;
   return result;
}

/** save a hdd to a imgage file */
void HddToImage(tCHS *chs,char *fname,int disk)
{
   FILE   *stream;
   long   sectors = 0;
   int    status;
   char   ch;
   if ((disk <0x80) || (disk > 0x83))
   {
      printf("disk number 0x%x not supported!\n",disk&0x7F);
      exit (0);
   }
   printf("read hdd%d to image <%s>\n",disk&0x7F,fname);
   printf("   using (%ld/%ld/%ld)\n",chs->max.Cylinder
                                    ,chs->max.Head
                                    ,chs->max.Sector);
   printf("continue (y/n)\n");
   ch = getch();
   if ((ch != 'y') && (ch != 'Y')) exit (0);
   stream = fopen(fname,"wb");
   if (stream)
   {
      printf("reading...");
      while(MakeChs(chs,sectors))
      {
         status = ReadSector(chs,RdBuffer,disk);
         if (status)
         {
            printf("\nread failed with code: 0x%x\n",status);
            printf("cyl    : %ld\nhead   : %ld\n\sector : %ld\n",
                    chs->cur.Cylinder,
                    chs->cur.Head,
                    chs->cur.Sector);
            fclose(stream);
            exit(0);
         }  
         else fwrite (RdBuffer, SECTORSIZE,1,stream);
         sectors++;
         if ((sectors % 1000) == 0) printf ("*");
      }
      fclose(stream);
      printf ("\nsuccess!\n");
   }
   else
   {
      printf ("error writing image file <%s>!\n",fname);
   }
}

/** write a image file to a hdd */
void ImageToHdd(tCHS *chs,char *fname,int disk)
{
   FILE   *stream;
   int    handle;
   int    status;
   long   fsize;
   long   maxsectors;
   long   sectors;
   char   ch;   
   if ((disk <0x80) || (disk > 0x83))
   {
      printf("disk number 0x%x not supported!\n",disk&0x7F);
      exit (0);
   }  
   
   /* calc the number of sectors */
   handle = open(fname,O_RDONLY);
   if (handle) 
   {
      fsize    = filelength(handle);
      maxsectors = fsize / SECTORSIZE;
      close (handle);
   }
   else
   {
      printf("file not found <%s>",fname);
      exit (0);
   }    
   
   printf("write image <%s> to hdd%d \n",fname,disk&0x7F);
   printf("   using (%ld/%ld/%ld)\n",chs->max.Cylinder
                                    ,chs->max.Head
                                    ,chs->max.Sector);
   printf("continue (y/n)");
   ch = getch();
   if ((ch != 'y') && (ch != 'Y')) exit (0);                              

   stream = fopen(fname,"rb");   
   if (stream)
   {  /* successfully opened */
      sectors = 0;
      while ((sectors < maxsectors) && (MakeChs(chs,sectors)))
      {  /* read from file and write to hdd */
         fread(WrBuffer, SECTORSIZE,1,stream);
         if (Test==0)
         {  
            status = WriteSector(chs,WrBuffer,disk);
            if (status)
            {  
               printf("\nwrite failed with code: 0x%x\n",status);
               printf("cyl    : %ld\nhead   : %ld\n\sector : %ld\n",
                       chs->cur.Cylinder,
                       chs->cur.Head,
                       chs->cur.Sector);
               PrintSector(WrBuffer,16);
               fclose(stream);
               exit(0);
            }
            /* dont use verify(AH=4) instead read the sector back */
            status = ReadSector(chs,RdBuffer,disk);
            if (status)
            {
               printf("\nreadback failed with code: 0x%x\n",status);
               printf("cyl    : %ld\nhead   : %ld\n\sector : %ld\n",
                       chs->cur.Cylinder,
                       chs->cur.Head,
                       chs->cur.Sector);
               fclose(stream);
               exit (0);
            }
            /* and compare both buffers */
            status = CompareBuffer(WrBuffer,RdBuffer,SECTORSIZE);
            if (status != -1)
            { /* there was some compare error */
               printf("\nverify failed at index [%d]\n",status);
               printf("wrbuffer : ");
               PrintSector(&WrBuffer[status / 0x10],1);
               printf("rdbuffer : ");
               PrintSector(&RdBuffer[status / 0x10],1);
               fclose(stream);
               exit (0);
            }
         }   
         sectors ++;
         /* do some output after 1000 sectors */
         if ((sectors % 1000) == 0) printf ("*");
      } /* while */
      fclose (stream);
      printf ("\nsuccess!\n");
   }
   else printf ("error opening image file <%s>!\n",fname);   
}

/** Make a one to one copy from one hdd to another hdd */
void DirectCopy(tCHS *chs,int srcdisk, int destdisk)
{
   int  status;
   long sectors = 0;
   char ch; 
   if ((srcdisk <0x80) || (srcdisk > 0x83))
   {
      printf("source disk number 0x%x not supported!\n",srcdisk);
      exit (0);
   }  
   
   if ((destdisk <0x80) || (destdisk > 0x83))
   {
      printf("destination disk number 0x%x not supported!\n",destdisk);
      exit (0);
   }  
   
   if (srcdisk ==destdisk)
   {
      printf("source:0x%x and destination:0x%x can't be same!\n",srcdisk,destdisk);
      exit (0);
   }  
   
   printf("copy hdd%d to hdd&d\n",srcdisk&0x7F,destdisk&0x7F);
   printf("   using (%ld/%ld/%ld)\n",chs->max.Cylinder
                                    ,chs->max.Head
                                    ,chs->max.Sector);
   printf("continue (y/n)");
   ch = getch();
   if ((ch !='y') && (ch != 'Y')) exit (0);                              
   
   while(MakeChs(chs,sectors))
   {  
      status = ReadSector (chs,RdBuffer,srcdisk);
      if (status)
      {
         printf("\nread failed with code: 0x%x\n",status);
         printf("cyl    : %ld\nhead   : %ld\n\sector : %ld\n",
                   chs->cur.Cylinder,
                   chs->cur.Head,
                   chs->cur.Sector);
         exit(0);
      } 
      status = WriteSector (chs,RdBuffer,destdisk);
      if (status)
      {
         printf("\nread failed with code: 0x%x\n",status);
         printf("cyl    : %ld\nhead   : %ld\n\sector : %ld\n",
                   chs->cur.Cylinder,
                   chs->cur.Head,
                   chs->cur.Sector);
         exit(0);
      }       
      sectors ++;
      if ((sectors % 1000) == 0) printf ("*");
   }  
}

/** scan possible drives */
void ScanDrives(char *buffer)
{
   int cyl;
   int head;
   int sect;
   int status;
   int i;
   
   for (i=0; i<4; i++)
   {
      status = biosdisk ( _DISK_INFO,     /* Int13 AH== 0x08 info */
                          i | 0x80,       /* drive    */
                          head,cyl,sect,  /* not used */
                          1, buffer);     /* 1 sector */
      if (status)
      { 
         printf("drive %d not found!\n",i & 0x7F);
      }
      else /* cl,ch, dl, dh */
      {
         sect   = buffer[0] & 0x63;
         cyl    = buffer[0] & 0xC0;
         cyl  <<= 2;
         cyl   |= buffer[1];
         head   = buffer[3];
         printf("drive %d found CHS",i & 0x7F);
         printf("(%d/%d/%d)\n",cyl,head,sect);
      }
   }
}

void DisplayHelp(void)
{
   printf("\noptions :\n");
   printf("   -d[0..3] : drive number                     (default: -d1)\n");
   printf("   -c????   : number of cylinder               (default: -c980)\n");
   printf("   -h????   : number of heads                  (default: -h15)\n");
   printf("   -s????   : number of sectors                (default: -s17)\n");
   printf("   -wIMAGE  : write image using CHS to HDD\n");
   printf("   -rIMAGE  : read HDD using CHS to image\n");
   printf("   -x[0..3] : copy from hdd -x? to the hdd -d?\n");
   printf("   -i       : scan drives \n\n");
   printf("example: imager -d0 -rK2000.IMG read from disk0 and write to <K2000.IMG>\n");
   printf("         imager -x0 -d2         read from disk0 and write to disk2\n");
   printf("         imager -i              scan drives \n");
}


void main(int argc, char *argv[])
{
   char  fname[256];
   char  option;
   int   i;
   tCHS  chs;
   int   src=-1;
   
   int   disk    = TARGETDISK;
   int   cyl     = MAXCYLINDER;
   int   head    = MAXHEAD;
   int   sect    = MAXSECTOR;
   int   mode    = 0;  /* no filename was given */
   
   printf("\imager: \n");
   /* collect the options */
   for (i=1;i < argc; i++)
   {
      if ((*argv[i]=='-') || (*argv[i]=='/'))
      {  /* all params start with - or / */
         argv[i]++;
         option = toupper (*argv[i]++);
         switch (option)
         {
            case 'D':
                 disk = strtol (argv[i],NULL,0);
                 disk |= 0x80;
                 break;
            case 'C':
                 cyl = strtol (argv[i],NULL,0);
                 break;
            case 'H':
                 head = strtol (argv[i],NULL,0);
                 break;
            case 'S':
                 sect = strtol (argv[i],NULL,0);
                 break;
            case 'R':
                 strcpy(fname,argv[i]);
                 mode = 1;
                 break;
            case 'W': 
                 strcpy(fname,argv[i]);
                 mode = 2;
                 break;
            case 'X': 
                 src = strtol (argv[i],NULL,0);
                 src |= 0x80;
                 mode = 3;
                 break;
            case 'I':
                 ScanDrives(RdBuffer);
                 exit(0);
                 break;       
            default:
                 printf("parameter not recognized -%c%s\n",option,argv[i]);
                 DisplayHelp();
                 exit(0);     
         }
      }
      else
      {  
         printf ("no parameter given!\n");
         DisplayHelp();  
         exit (0);
      } 
   }
   /* save the CHS parameter */
   chs.max.Cylinder = cyl;
   chs.max.Head     = head;
   chs.max.Sector   = sect;

   switch (mode)
   {  /* mode 1 ..3*/
      case 0:
           printf ("use -R or -W to set a filename!\n");
           DisplayHelp();
           exit (0);
           break;
      case 1:
           HddToImage(&chs,fname,disk);
           break;
      case 2:
           ImageToHdd(&chs,fname,disk);
           break;
      case 3:
           DirectCopy(&chs,src,disk);
           break;
      default:
           printf ("program error don't know what to do!\n");
   }
}   
